#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id$
# Copyright (C) 2008, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

"""
BitTorrent distribution tool client
"""

import os
import os.path
import re
import sys
import time

from btdt.bencode import bdecode
from btdt.common import AuthorizedServerProxy, config_default_values, \
                        size_to_human_readable, mkstemp
from btdt.logger import init_logger
from btdt.tracker_proxy import TrackerProxy
from ConfigParser import ConfigParser
from optparse import OptionParser
from threading import Thread

from btdt.deluge_wrapper import DelugeManager, \
    default_config_parser_options as deluge_default_config_parser_options

__version__ = '0.1'


def choose_file(server):
    files = server.list_files()
    if not len(files):
        print 'Keine Dateien verfuegbar!'
        return None

    print 'Verfuegbare Dateien:'
    while True:
        for i, (name, torrents_data) in enumerate(files):
            info = bdecode(torrents_data.data)['info']
            if 'files' in info:
                size = sum((f['length'] for f in info['files']))
            else:
                size = info['length']
            print '  %i) %s (%s)' % (
                i + 1,
                name.encode(sys.stdout.encoding, 'replace'),
                size_to_human_readable(size)
            )
        try:
            i = raw_input('Datei auswaehlen (a = alle): ')
        except (EOFError, KeyboardInterrupt):
            return None
        if i == 'a':
            return files

        l = i.split(' ')
        valid = True
        for t in l:
            if not t.isdigit():
                valid = False
                break

        if not valid:
            continue

        l = [int(t) - 1 for t in l if int(t) > 0 and int(t) <= len(files)]
        return [files[i] for i in l]

def check_host_from_cmdline(server, port=None, protocol='https'):
    m = re.match('^[a-z]+://[a-zA-Z0-9_.-]+(:\d{1,5})?/?\Z', server)
    if m:
        return server

    # no port?
    if port is not None:
        m = re.match('^[^/]*:\d{1,5}/?\Z', server)
        if not m:
            server = '%s:%d' % (server, port)

    # no protocol?
    m = re.match('^[a-z]+://.*', server)
    if m:
        # something else is wrong
        return None
    else:
        server = '%s://%s' % (protocol, server)

    return check_host_from_cmdline(server)

def main():
    config = ConfigParser()
    config.read('config.ini')

    # default values
    default_values = {
        'Client': {
            'Host': '',
            'Minport': 16000,
            'Maxport': 16009,
            'Maxupload': 0,
            'StorageMode': 1,
            'DefaultRPCPort': 15002,
            'DefaultRPCProtocol': 'https',
            # Authorization
            'Username': '',
            'Password': ''
        },
        'ClientLogging': {
            'File': 'btdt-client.log',
            'Level': 'DEBUG',
            'StdoutLevel': 'WARNING'
        }
    }
    config_default_values(config, default_values)

    usage = 'usage: %prog [options] /out/dir'
    version = '%prog version ' + __version__

    parser = OptionParser(usage=usage, version=version)
    parser.add_option('--minport', dest='minport', type='int',
                      default=config.getint('Client', 'Minport'))
    parser.add_option('--maxport', dest='maxport', type='int',
                      default=config.getint('Client', 'Maxport'))
    parser.add_option('--maxupload', dest='maxupload', type='int',
                      default=config.getint('Client', 'Maxupload'),
                      help='maxupload in KB/s (0 = no limit)')
    parser.add_option('--update-interval', dest='update_interval', type='int',
                      default=2)
    parser.add_option('--storage-mode', dest='storage_mode', type='int',
                      default=config.getint('Client', 'StorageMode'),
                      help='0 = sparse, 1 = compact')
    parser.add_option('-m', '--mirror', dest='mirror', action='store_true',
                      default=False)
    parser.add_option('--logfile', dest='logfile', type='string',
                      default=config.get('ClientLogging', 'File'),
                      help='log file')
    parser.add_option('--loglevel', dest='loglevel', type="string",
                      default=config.get('ClientLogging', 'Level'),
                      help='log file log level (one of DEBUG, INFO, WARNING, '
                      'ERROR, CRITICAL)')
    parser.add_option('--stdoutloglevel', dest='stdoutloglevel', type='string',
                      default=config.get('ClientLogging', 'StdoutLevel'),
                      help='log file log level (one of DEBUG, INFO, WARNING, '
                      'ERROR, CRITICAL)')
    parser.add_option('--disable-logging', dest='logfile', action='store_const',
                      const=None, help='disables logging')
    parser.add_option('-H', '--host', dest='host', type='string',
                      default=config.get('Client', 'Host'),
                      help='host to connect to')
    deluge_default_config_parser_options(config, parser)

    opts, args = parser.parse_args()
    if not len(args) == 1:
        parser.print_usage()
        return 1
    init_logger(opts.logfile, opts.loglevel, opts.stdoutloglevel)

    out_dir = args[0]
    host = check_host_from_cmdline(opts.host,
                                   config.getint('Client', 'DefaultRPCPort'),
                                   config.get('Client', 'DefaultRPCProtocol'))
    if host is None:
        print "'%s' is not a valid server!" % args[0]
        print "Expecting something like [protocol://]hostname[:port]."
        return 1

    username = config.get('Client', 'Username')
    password = config.get('Client', 'Password')
    server = AuthorizedServerProxy(host, username, password)
    if opts.mirror:
        print 'downloading all torrents ...'
        files = server.list_files()
    else:
        files = choose_file(server)

    if not files:
        return 1

    tracker_proxy = TrackerProxy(-1, host, username, password)
    proxy_thread = Thread(target=tracker_proxy.serve_forever,
                          name='btdt tracker proxy')
    proxy_thread.start()

    dm = DelugeManager([opts.minport, opts.maxport], maxupload=opts.maxupload,
                       use_utpex=opts.use_utpex, use_natpmp=opts.use_natpmp,
                       use_lsd=opts.use_lsd, use_upnp=opts.use_upnp,
                       fast_resume_path=config.get('Deluge', 'FastResumePath'),
                       webseed_workaround=config.getint('Deluge',
                                                        'WebseedWorkaround'))

    for f, torrent in files:
        # get the torrent ...
        torrent = torrent.data
        f, path = mkstemp()
        os.write(f, torrent)
        os.close(f)

        # ... and add it
        id_ = dm.add_torrent(path, out_dir, opts.storage_mode)
        dm.replace_trackers(
            id_,
            'http://localhost:%i/announce' % tracker_proxy.server_port
        )

    try:
        while True:
            for i in dm.ids():
                state = dm.get_torrent_state(i)
                print '%d: %s (%s, %.2f %%)' % (i + 1, state['name'],
                                                state['state_string'],
                                                state['progress_%'])
                print 'Down/Up: %.2f KB/s / %.2f KB/s' % \
                    (state['download_rate_kbs'], state['upload_rate_kbs'])
                print ''

            time.sleep(opts.update_interval)
    except KeyboardInterrupt:
        pass

    print 'Beende ...'
    dm.quit()
    tracker_proxy.shutdown()
    proxy_thread.join()

    return 0
