#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id$
# Copyright (C) 2008, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" some common functions
"""

from __future__ import with_statement

import atexit
import os
import os.path
import random
import shutil
import socket
import tempfile

from base64 import b64decode
from btdt.bencode import bencode, bdecode
from hashlib import sha256, sha1 as sha
from secure_server import SecureXMLRPCRequestHandler
from urllib import unquote
from urlparse import urlparse, urlsplit, urlunsplit
from xmlrpclib import SafeTransport, ServerProxy

class DefaultingDict(dict):
    """ 'defaulting' dict

    Behaves like dict but if an item in __getitem__ can't be found it will
    default initialized with parameter passed to __init__.
    """

    def __init__(self, default=None, *args, **kwargs):
        self.__default = default
        dict.__init__(self, *args, **kwargs)

    def __getitem__(self, item):
        if item not in self:
            self[item] = self.__default
        return dict.__getitem__(self, item)


class AuthorizedServerProxy(ServerProxy):
    """ Authorized server proxy

    AuthorizedServerProxy requires a username and password and is able to
    communicate with XML-RPC servers which require authorization.
    See xmlrpc.ServerProxy' documentation for further details.
    AuthorizedServerProxy("protocoll://url", "user", "password") is
    equivalent to xmlrpclib.ServerProxy("protocoll://user:password@url").
    """

    def __init__(self, url, username, password, *args, **kwargs):
        url = list(urlsplit(url))
        url[1] = '%s:%s@%s' % (username, password, url[1])
        ServerProxy.__init__(self, urlunsplit(url), *args, **kwargs)

class AuthorizedXMLRPCRequestHandler(SecureXMLRPCRequestHandler):
    """ Authorized XML-RPC request handler

    AuthorizedXMLRPCRequestHandler extends a XML-RPC-server with some basic
    authorization functionality. It requires the client to pass username and
    password and it calls the server's is_allowed method to check if the client
    provided a valid username and password.
    """

    def do_POST(self):
        # Only authorized users are allowed to use the server
        if not 'Authorization' in self.headers:
           return self.send_error(401)
        else:
            try:
                auth_header = self.headers['authorization']
                type_, auth_data = auth_header.split()
                if type_ != 'Basic':
                    raise ValueError()
                username, password = b64decode(auth_data).split(':', 1)
                password = password_hash(password)
                if not self.server.is_allowed(username, password):
                    raise ValueError()
            except (TypeError, ValueError):
                return self.send_error(403)

        return SecureXMLRPCRequestHandler.do_POST(self)

def path_size(f):
    """
    Recursively calculates the size of the given path and all its'
    subdirectories and files if it's a directory.
    """

    if os.path.isdir(f):
        return sum(path_size(os.path.join(f, t)) for t in os.listdir(f))
    else:
        return os.stat(f).st_size

def remove_path(path):
    """ Removes a file/directory. """
    if os.path.isdir(path):
        shutil.rmtree(path)
    else:
        os.unlink(path)

def config_default_values(config, default_values):
    for section in default_values.iterkeys():
        if not config.has_section(section):
            config.add_section(section)

        for opt, val in default_values[section].iteritems():
            if not config.has_option(section, opt):
                config.set(section, opt, str(val))

def config_section_as_dict(config, section):
    d = {}
    for opt, val in config.items(section):
        d[opt] = val
    return d

def delete_torrent_fastresume(path):
    """ Deletes a torrent file and its fast resume data. """

    fastresume = '%s.fastresume' % path
    if os.path.exists(path):
        os.unlink(path)
    if os.path.exists(fastresume):
        os.unlink(fastresume)

def read_torrent(path):
    """
    Opens a torrent file and returns its data, info hash and the info hash
    in hex.
    """

    with open(path, 'rb') as fp:
        data = bdecode(fp.read())

    s = sha(bencode(data['info']))
    info_hash = s.digest()
    info_hash_hex = s.hexdigest()

    return (data, info_hash, info_hash_hex)

def mkstemp(*args, **kwargs):
    """
    Wraps around tempfile.mkstemp and deletes the temporary file at exit.
    """

    f, path = tempfile.mkstemp(*args, **kwargs)
    atexit.register(os.unlink, path)

    return (f, path)

def _flatten(lol):
    for i in lol:
        if isinstance(i, (tuple, list)):
            for t in _flatten(i):
                yield t
        else:
            yield i

def flatten(lol):
    """ Flattens tuples and lists. """

    return list(_flatten(lol))

UQ_BROKEN = unquote('+') != ' '

def furlparse(path):
    """ Wraps urlparse.urlparse. """

    (scheme, netloc, path, pars, query, fragment) = urlparse(path)
    if UQ_BROKEN:
        path = path.replace('+',' ')
        query = query.replace('+',' ')

    return (scheme, netloc, path, pars, query, fragment)

def random_string(length=32, chars="abcdefghijklmnopqrstuvwxyz" \
                                   "ABCDEFGHIJKLMNOPQRSTUVWXYZ" \
                                   "0123456789_"):
    """ Generates a random string. """

    return "".join([random.choice(chars) for i in xrange(length)])

def password_hash(pw, fn=sha256, ashex=True):
    """ Calculates password hash. """

    return fn(pw).hexdigest() if ashex else fn(pw).digest()

def size_to_human_readable(size):
    """ Convertes size into a human readable format. """

    for name in ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB']:
        if size < 1024:
            return '%3.1f %s' % (size, name)
        size /= 1024.0
    return '%i %s' % (size, name)
