#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id$
# Copyright (C) 2008, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" Deluge wrapper

Wraps some functions from deluge.deluge_core.
"""

from __future__ import with_statement

from btdt.bencode import bencode, bdecode
from btdt.common import config_default_values, size_to_human_readable, \
                        read_torrent
from btdt.logger import logger
from tempfile import mktemp
from threading import Event, Thread

import deluge.deluge_core as dc
import os.path
import re

STATUS_STRING = (
    'queued for checking',
    'checking files',
    'connecting to tracker',
    'downloading metadata',
    'downloading',
    'finished',
    'seeding',
    'allocation'
)

FORCE_ENC = 0
ENABLE_ENC = 1
DISABLE_ENC = 2
PLAIN_ENC = 0
RC4_ENC = 1
BOTH_ENC = 2

ENC_MODES = (FORCE_ENC, ENABLE_ENC, DISABLE_ENC)
ENC_LEVEL_MODES = (PLAIN_ENC, RC4_ENC, BOTH_ENC)

DEFAULT_CONFIG = {
    'UseUPnP': 0,
    'UsePeerEx': 1,
    'UseDHT': 0,
    'UseLSD': 0,
    'UseNatPMP': 0,
    'FastResumePath': '',
    'WebseedWorkaround': 1
}

# exceptions from deluge's core.py

class DelugeError(Exception):
    """ DelugeError from deluge.core """

    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class InvalidEncodingError(DelugeError):
    """ InvalidEncodingError from deluge.core """

class FilesystemError(DelugeError):
    """ FilesystemError from deluge.core """

class DuplicateTorrentError(DelugeError):
    """ DuplicateTorrentError from deluge.core """

class InvalidTorrentError(DelugeError):
    """ InvalidTorrentError from deluge.core """

class InvalidUniqueIDError(DelugeError):
    """ InvalidUniqueIDError from deluge.core """

class DelugeManager(object):
    """ Deluge manager

    Manages the torrents and some deluge related settings.
    """

    use_upnp = dc.use_upnp
    use_natpmp = dc.use_natpmp
    use_lsd = dc.use_lsd
    use_utpex = dc.use_utpex

    def __init__(self, ports,
                 pe_settings=(FORCE_ENC, FORCE_ENC, BOTH_ENC, True),
                 maxupload=0, use_upnp=False, use_lsd=False, use_natpmp=False,
                 use_utpex=False, fast_resume_path=None,
                 webseed_workaround=0):
        """ Initializes deluge's core and sets some options """

        dc.pre_init(DelugeError, InvalidEncodingError, SystemError,
                    FilesystemError, DuplicateTorrentError, InvalidTorrentError)
        dc.init("LT", 0, 13, 0, 0, "BTDT")

        self.pe_settings(*pe_settings)
        self.listen(*ports)
        self.upload_rate_limit(maxupload)

        # dc.stop_DHT()
        self.use_upnp(use_upnp)
        self.use_lsd(use_lsd)
        self.use_natpmp(use_natpmp)
        self.use_utpex(use_utpex)

        self.fast_resume_path = None
        if fast_resume_path is not None and len(fast_resume_path):
            if not os.path.exists(fast_resume_path):
                os.mkdir(fast_resume_path)
            if os.path.isdir(fast_resume_path):
                self.fast_resume_path = fast_resume_path

        self.torrents = {}

        self.webseed_thread = None
        if webseed_workaround:
            self.webseed_workaround_interval = webseed_workaround * 60
            self.doneevent = Event()
            self.webseed_thread = Thread(target=self.__webseed_workaround,
                                         name='btdt DM webseed bug fix')
            self.webseed_thread.start()

    def quit(self, fastresume=True):
        """ Deinitializes deluge's core and saves fast resume data. """

        self.doneevent.set()
        self.webseed_thread.join()

        # save resume data
        if fastresume:
            for id in tuple(self.ids()):
                self.remove_torrent(id, delete_data=False)

        dc.quit()

    def listen(self, minport, maxport=None):
        """
        Tells Deluge to listen on one of the ports in [minport, maxport] or on
        minport if maxport is None.
        """

        if maxport is None:
            ports = [minport, minport]
        else:
            ports = sorted([minport, maxport])

        logger().debug("listening on %s" % str(ports))
        dc.set_listen_on(ports)

    def listening_info(self):
        """
        Retuns False if Deluge is not listening on any port or the open port
        otherwise.
        """

        if dc.is_listening():
            return dc.listening_port()
        else:
            return False

    def pe_settings(self, encout=FORCE_ENC, encin=FORCE_ENC, level=BOTH_ENC,
                    prefer_rc4=True):
        if encout not in ENC_MODES or \
            encin not in ENC_MODES or \
            level not in ENC_LEVEL_MODES:
            raise ValueError('invalid encryption settings')

        dc.pe_settings(encout, encin, level, prefer_rc4)

    def upload_rate_limit(self, limit):
        """
        Sets the upload rate limit (in KB/s). The upload rate is unlimited if
        the argument is smaller or equal to 0.
        """

        if limit <= 0:
            limit = -1
        else:
            limit = limit * 1024

        if limit != -1:
            logger().debug("setting upload rate limit to %s/s" % \
                           size_to_human_readable(limit))
        else:
            logger().debug("setting upload rate limit to unlimited")
        dc.set_upload_rate_limit(limit)

    def add_torrent(self, path, target, compact=False):
        """ Adds a torrent to the download queue. """

        if not os.path.exists(path):
            raise ValueError('torrent path doest not exist')
        logger().debug("adding torrent '%s'" % path)

        info_hash_hex = read_torrent(path)[2]
        if self.fast_resume_path is not None:
            fsd = os.path.join(self.fast_resume_path,
                               '%s.fastresume' % info_hash_hex)
            if os.path.exists(fsd):
                logger().debug("found fast resume data for '%s'" % path)
                with open("%s.fastresume" % path, 'wb') as fo:
                    with open(fsd, 'rb') as fi:
                        fo.write(fi.read())

        id = dc.add_torrent(path, target, compact)
        self.torrents[id] = {
            'path': path,
            'target': target,
            'infohash': info_hash_hex
        }

        state = self.get_torrent_state(id)
        self.torrents[id]['name'] = state['name']

        if self.webseed_thread is not None:
            data = read_torrent(path)[0]
            if 'url-list' in data and len(data['url-list']):
                self.torrents[id]['urls'] = data['url-list']

        return id

    def pause_torrent(self, id):
        """ Pauses the download of a torrent. """

        id = self.check_or_find(id)
        state = self.get_torrent_state(id)
        if not state['is_paused']:
            dc.pause(id)

    def remove_torrent(self, id, delete_data=False):
        """ Removes a torrent from the download queue. """

        id = self.check_or_find(id)
        if self.fast_resume_path is not None:
            info_hash_hex = self.torrents[id]['infohash']
            # save_fastresume adds .fastresume
            fsd = os.path.join(self.fast_resume_path, info_hash_hex)
            fsd2 = '%s.fastresume' % fsd

            if not delete_data:
                path = self.torrents[id]['path']
                logger().debug("saving fastresume data for '%s'" % path)
                dc.save_fastresume(id, fsd)

                # paranoid mode
                with open(fsd2, 'r') as f:
                    data = bdecode(f.read())
                data['peers'] = {}
                with open(fsd2, 'w') as f:
                    f.write(bencode(data))
            elif os.path.exists(fsd2):
                os.unlink(fsd2)

        dc.remove_torrent(id, delete_data)
        del self.torrents[id]

    def resume_torrent(self, id):
        """ Resumes the download of a torrent. """

        id = self.check_or_find(id)
        state = self.get_torrent_state(id)
        if state['is_paused']:
            dc.resume(id)

    def get_peer_info(self, id_):
        """ Retrieves some information on a torrent's peers. """
        id_ = self.check_or_find(id_)
        return dc.get_peer_info(id_)

    def get_torrent_state(self, id):
        """ Retrieves some information on a torrent's state. """

        id = self.check_or_find(id)
        state = dc.get_torrent_state(id)
        state['state_string'] = STATUS_STRING[state['state']]
        state['progress_%'] = state['progress'] * 100.0
        state['download_rate_kbs'] = state['download_rate'] / 1024.0
        state['upload_rate_kbs'] = state['upload_rate'] / 1024.0

        # Calculate ETA
        size = state['total_wanted']
        done = state['total_wanted_done']
        speed = int(state['download_rate'])
        if (size - done) == 0 or speed == 0:
            eta = 0
        else:
            eta = (size - done) / speed
        state['eta'] = eta

        return state

    def get_state(self):
        """ Retrives some information on the session's state. """

        state = dc.get_session_info()
        state['download_rate_kbs'] = state['download_rate'] / 1024.0
        state['upload_rate_kbs'] = state['upload_rate'] / 1024.0
        return state

    def ids(self):
        """ Retrives the torrent's IDs. """

        return self.torrents.iterkeys()

    def check_id(self, id):
        """ Checks if the id exists. """

        return id in self.torrents

    def find_id(self, torrent):
        """ Finds the unique id based on the torrent's name. """

        for tor in self.torrents.itervalues():
            if tor['name'] == torrent:
                return id

        return None

    def check_or_find(self, id):
        """
        Checks if the id exists or if there's a torrent with that name. Returns
        the id on success and raises InvalidUniqueIDError on failure.
        """

        if self.check_id(id):
            return id

        nid = self.find_id(id)
        if nid is not None:
            return nid

        raise InvalidUniqueIDError("torrent with name or id '%s' doesn't exist"
                                   % str(id))

    def replace_trackers(self, id, trackers):
        """ Replaces the tracker of a specific torrent. """

        id = self.check_or_find(id)
        return dc.replace_trackers(id, trackers)

    def get_webseeds(self, id):
        """ Retrieve list of webseeds. """

        peers = self.get_peer_info(id)
        webseeds = []
        for peer in peers:
            # web seeds are seeding
            if not peer['is_seed']:
                continue

            res = re.search('URL seed @ ([A-Za-z.:0-9-]+) \(.+\)',
                            peer['client'])
            if res is not None:
                webseeds.append(res.group(1))

        return webseeds

    def __webseed_workaround(self):
        while not self.doneevent.isSet():
            for id, info in self.torrents.iteritems():
                if 'urls' not in info:
                    continue

                state = self.get_torrent_state(id)
                if state['is_paused'] or state['state_string'] != 'downloading':
                    continue

                webseeds = self.get_webseeds(id)
                urls = info['urls']
                if isinstance(urls, basestring):
                    urls = (urls, )

                logger().debug("webseeds are '%s', should be '%s'" % \
                               (str(webseeds), str(urls)))
                for url in urls:
                    found = False
                    for webseed in webseeds:
                        if webseed in url:
                            found = True

                    if not found:
                        logger().info("readding webseed '%s'" % url)
                        dc.add_url_seed(id, url)

            self.doneevent.wait(self.webseed_workaround_interval)


def _d057_create_torrent_workaround(*args):
    """ Workaround for Deluge < 0.5.8. """

    dc.create_torrent(*args[:7])

    destination = args[0]
    webseeds = args[7]

    if len(webseeds):
        data = read_torrent(destination)[0]

        data['url-list'] = webseeds.split('\n')
        with open(destination, 'wb') as f:
            f.write(bencode(data))

_ct = dc.create_torrent
def create_torrent(destination, input, trackers, comment='', piece_size=1024,
                   creator='BTDT', priv=False, webseeds=''):
    """
    Creates a torrent and takes the same arguments as deluge.deluge_core's
    create_torrent.
    """

    if not os.path.exists(input):
        raise ValueError("input path doesn't exist")

    dest = mktemp()
    # FIXME
    t = trackers.split('\n')
    args = (dest, input, trackers, comment, piece_size, creator, priv,
            webseeds)

    global _ct
    try:
        _ct(*args)
    except TypeError:
        _ct = _d057_create_torrent_workaround
        _ct(*args)

    with open(destination, 'wb') as fo:
        with open(dest, 'rb') as fi:
            fo.write(fi.read())

def default_config_parser_options(config, parser):
    """
    Adds some Deluge specific options to the passed optparse.OptionParser
    instance.
    """

    config_default_values(config, {'Deluge': DEFAULT_CONFIG})

    parser.add_option('--peer-exchange', dest='use_utpex', action='store_true',
                      default=config.getint('Deluge', 'UsePeerEx'))
    parser.add_option('--no-peer-exchange', dest='use_utpex',
                      action='store_false')

    parser.add_option('--upnp', dest='use_upnp', action='store_true',
                      default=config.getint('Deluge', 'UseUPnP'))
    parser.add_option('--no-upnp', dest='use_upnp', action='store_false')

    parser.add_option('--dht', dest='use_dht', action='store_true',
                      default=config.getint('Deluge', 'UseDHT'))
    parser.add_option('--no-dht', dest='use_dht', action='store_false')

    parser.add_option('--nat-pmp', dest='use_natpmp', action='store_true',
                      default=config.getint('Deluge', 'UseNatPMP'))
    parser.add_option('--no-nat-pmp', dest='use_natpmp', action='store_false')

    parser.add_option('--lsd', dest='use_lsd', action='store_true',
                      default=config.getint('Deluge', 'UseLSD'))
    parser.add_option('--no-lsd', dest='use_lsd', action='store_false')
