#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id$
# Copyright (C) 2008, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" BitTorrent distribution tool server

Provides the server's main function.
"""

import os.path
import sys
import time

from btdt.admin_rpc import AdminRPC
from btdt.common import config_default_values, config_section_as_dict, \
                        password_hash
from btdt.custom_tracker import Tracker
from btdt.database import init_db
from btdt.logger import init_logger, logger
from btdt.secure_server import create_certificate
from btdt.tracker_rpc import TrackerRPC
from btdt.webseeder import Webseeder
from ConfigParser import ConfigParser
from optparse import OptionParser
from threading import Thread

from btdt.deluge_wrapper import DelugeManager, \
    default_config_parser_options as deluge_default_config_parser_options

# default values
default_values = {
    'Tracker': {
        'Host': '',
        'Port': 15001,
        'Torrents_Path': 'torrents',
        'Data_Path': 'torrents-data',

        # some more options
        'Timeout_Downloaders_Interval': 45 * 60,  # seconds between expiring downloaders
        'Reannounce_Interval': 5 * 60,  # seconds downloaders should wait between reannouncements
        'Response_Size': 50, # number of peers to send in an info message
        'NAT_Check': 3,  # how many times to check if a downloader is behind a NAT (0 = don't check)
        'Parse_Allowed_Interval': 15, # minutes between reloading of allowed_dir
        'Show_Names': 1, # whether to display names from allowed dir
        'Only_Local_Override_IP': 1,   # ignore the ip GET parameter from machines which aren't on local network IPs
        'Max_Give': 200  # maximum number of peers to give with any one request
    },
    'TrackerRPC': {
        'Host': '',
        'Port': 15002
    },
    'Uploader': {
        'Port': 15003,
        'Maxupload': 0,
        'Max_Handler_Threads': 10
    },
    'SSL': {
        'Key': 'key.pem',
        'Cert': 'cert.pem'
    },
    'RemoteAdminLogin': {
        'Enabled': 0,
        'Port': 15004
    },
    'Database': {
        'URL': 'sqlite:///database.sqlite'
    },
    'Logging': {
        'File': 'btdt-server.log',
        'Level': 'DEBUG',
        'StdoutLevel': 'WARNING'
    }
}

def main():
    """ Runs the server. """
    config = ConfigParser()
    config.read('config.ini')
    config_default_values(config, default_values)

    usage = 'usage: %prog [options]'
    version = '%prog version 1.0'

    parser = OptionParser(usage=usage, version=version)
    parser.add_option('--maxupload', dest='maxupload', type='int',
                      default=config.getint('Uploader', 'Maxupload'),
                      help='maxupload in KB/s (0 = no limit)')
    parser.add_option('--logfile', dest='logfile', type='string',
                      default=config.get('Logging', 'File'), help='log file')
    parser.add_option('--loglevel', dest='loglevel', type="string",
                      default=config.get('Logging', 'Level'),
                      help='log file log level (one of DEBUG, INFO, WARNING, '
                      'ERROR, CRITICAL)')
    parser.add_option('--stdoutloglevel', dest='stdoutloglevel', type='string',
                      default=config.get('Logging', 'StdoutLevel'),
                      help='log file log level (one of DEBUG, INFO, WARNING, '
                      'ERROR, CRITICAL)')
    parser.add_option('--disable-logging', dest='logfile', action='store_const',
                      const=None, help='disables logging')

    opts, args = parser.parse_args()
    if len(args):
        parser.print_usage()
        return 1

    init_logger(opts.logfile, opts.loglevel, opts.stdoutloglevel)
    log = logger()
    log.info('===== starting log =====')

    # initialize server
    log.info('initializing btdt-server ...')

    # check existence of key and certificate
    key_file = config.get('SSL', 'Key')
    cert_file = config.get('SSL', 'Cert')
    if not os.path.exists(key_file) or not os.path.exists(cert_file):
        log.info('creating key and certificate ...')
        create_certificate(config.get('Tracker', 'Host'), key_file, cert_file)

    webseed_port = config.getint('Uploader', 'Port')
    rpc_host = config.get('TrackerRPC', 'Host')
    rpc_port = config.getint('TrackerRPC', 'Port')
    tracker_host = config.get('Tracker', 'Host')
    tracker_port = config.getint('Tracker', 'Port')
    tracker_config = config_section_as_dict(config, 'Tracker')
    admin_enabled = config.getint('RemoteAdminLogin', 'Enabled')
    admin_port = config.getint('RemoteAdminLogin', 'Port')
    database_url = config.get('Database', 'URL')

    # initalize database
    log.info('initializing database ...')
    init_db(database_url, echo=False)

    # initialize services
    log.info('initializing services ...')
    tracker = Tracker(tracker_config, (tracker_host, tracker_port),
                      webseed_port, key_file, cert_file)
    rpc = TrackerRPC((rpc_host, rpc_port), tracker, key_file, cert_file)
    webseed = Webseeder(webseed_port, tracker,
                        config.getint('Uploader', 'Max_handler_threads'),
                        opts.maxupload)
    admin = AdminRPC(('', admin_port), tracker, rpc, key_file, cert_file,
                     admin_enabled)

    # start threads
    log.info('creating TrackerRPC thread ...')
    rpc_thread = Thread(target=rpc.serve_forever, name='btdt tracker RPC')
    rpc_thread.start()

    log.info('creating Webseeder thread ...')
    webseed_thread = Thread(target=webseed.serve_forever,
                            name='btdt webseed')
    webseed_thread.start()

    if admin_enabled:
        log.info('creating AdminRPC thread ...')
        admin_thread = Thread(target=admin.serve_forever,
                              name='btdt admin RPC')
        admin_thread.start()

    log.info('btdt-server running')

    def print_(f):
        def impl(*args):
            res = f(*args)
            print res
            return res

        return impl

    def print_help(*args):
        h = (
            ('h(elp)', "prints this list"),
            ('q(quit)', "terminates the server"),
            ('adduser name password', "adds a new user or updates the user's password"),
            ('deleteruser name', "deletes an existing user"),
            ('listusers', "lists all users"),
            ('addadmin name', "gives admin rights to an existing user"),
            ('deleteadmin name', "removes admin rights from an existing user"),
            ('listadmins', "lists all admins"),
            ('u(pdate)', "updates the list of torrents")
        )

        for t in h:
            print '%s:\t%s' % t

    cmds = {
        'help': print_help,
        'quit': AdminRPC.shutdown_server,
        'adduser': AdminRPC.add_user,
        'deleteuser': AdminRPC.delete_user,
        'listusers': print_(AdminRPC.list_users),
        'addadmin': AdminRPC.add_admin,
        'deleteadmin': AdminRPC.delete_admin,
        'listadmins': print_(AdminRPC.list_admins),
        'update': AdminRPC.update_torrents
    }

    scmds = {
        'q': 'quit',
        'h': 'help',
        'u': 'update'
    }

    # handle user input
    try:
        while True:
            t = raw_input('> ')
            cmdline = t.split(' ')
            if not len(cmdline):
                continue

            cmd = cmdline[0]
            args = cmdline[1:]

            try:
                if cmd in cmds:
                    cmds[cmd](admin, *args)
                elif cmd in scmds:
                    cmds[scmds[cmd]](admin, *args)
                else:
                    log.error("command '%s' unknown (try help)" % cmd)
            except (KeyboardInterrupt, EOFError):
                raise
            except Exception, e:
                log.error("'%s' failed (exception follows):" % cmd)
                log.exception(e)
    except (KeyboardInterrupt, EOFError):
        pass

    # shutdown

    log.info('shutting down ...')
    if admin_enabled:
        admin.shutdown()
    webseed.shutdown()
    rpc.shutdown()
    tracker.shutdown()

    log.info('joining threads ...')
    if admin_enabled:
        admin_thread.join()
    webseed_thread.join()
    rpc_thread.join()

    log.info('===== bye bye =====')
    return 0
