#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id$
# Copyright (C) 2008, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" Tracker XML-RPC server

Provides the tracker's XML-RPC server.
"""

from btdt.secure_server import SecureXMLRPCServer
from btdt.common import AuthorizedXMLRPCRequestHandler, password_hash, \
                        AuthorizedServerProxy, random_string
from btdt.database import User, Session
from btdt.logger import logger
from threading import Event
from xmlrpclib import Binary
import sys
import socket

class TrackerRPCRequestHandler(AuthorizedXMLRPCRequestHandler):
    """ Tracker XML-RPC request handler

    Just extends AuthorizedXMLRPCRequestHandler with custom log_output.
    """

    def log_message(self, fmt, *args):
        logger().info('%s: %s' % (self.address_string(), fmt % args))

class TrackerRPC(SecureXMLRPCServer):
    """ Tracker XML-RPC server """

    allow_reuse_address = 1
    functions = list()

    def register(method, functions=functions):
        functions.append(method.__name__)
        return method

    def is_allowed(self, user, pw):
        """ Returns True if the user and password are valid. """

        if user == self.__quit_user and pw == password_hash(self.__quit_pw):
            return True

        return self.session.query(User).filter_by(name=user, password_hash=pw)\
            .count() == 1

    def __init__(self, host_port, tracker, key_file, cert_file):
        """ Initializes the XML-RPC server. """

        logger().info('initializing tracker XML-RPC server ...')

        self.doneevent = Event()
        self.host_port = host_port
        self.tracker = tracker
        self.session = Session()

        self.__quit_user = random_string(64)
        self.__quit_pw = random_string(64)

        SecureXMLRPCServer.__init__(self, host_port, key_file, cert_file,
                                    logRequests=True,
                                    HandlerClass=TrackerRPCRequestHandler)

        for name in self.functions:
            self.register_function(getattr(self, name))

    def get_request(self):
        req, addr = SecureXMLRPCServer.get_request(self)
        self.client_address = addr

        return (req, addr)

    def serve_forever(self):
        logger().info('tracker XML-RPC server starts serving ...')
        while not self.doneevent.isSet():
            try:
                self.handle_request()
            except socket.error, e:
                logger().exception(e)

    def shutdown(self):
        logger().info('shutting down tracker XML-RPC server ...')

        host, port = self.host_port
        if not host:
            host = 'localhost'
        p = AuthorizedServerProxy('https://%s:%d' % (host, port),
                                  self.__quit_user, self.__quit_pw)
        try:
            p.quit()
        except socket.error, e:
            logger().exception(e)

        logger().info('tracker XML-RPC server shut down done')

    @register
    def quit(self):
        self.doneevent.set()
        # just handle the request and do nothing
        return True

    @register
    def list_files(self):
        files = list()
        fsencoding = sys.getfilesystemencoding()
        for info in self.tracker.allowed.itervalues():
            files.append((info['name'].decode(fsencoding),
                          Binary(open(info['path'], 'rb').read())))
        return files

    @register
    def scrape(self, infohash, *args):
        code, header, data = self.tracker.scrape(infohash.data, *args)
        return code, header, Binary(data)

    @register
    def announce(self, *args):
        code, header, data = self.tracker.announce(self.client_address, *args)
        return code, header, Binary(data)
